"
 COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass:#ObjectFileLoader
       instanceVariableNames:''
       classVariableNames:'mySymbolTable stubNr verbose'
       poolDictionaries:''
       category:'System-Compiler'
!

ObjectFileLoader comment:'

COPYRIGHT (c) 1993 by Claus Gittinger
             All Rights Reserved

this one knowns how to load in external (c)-modules
(see fileIn/cExample.c) it is all experimental and 
WILL DEFINITELY change soon ...

(goal is to allow loading of binary classes)

%W% %E%
'!

%{
#ifdef NeXT
# ifndef _RLD_H_
#  define _RLD_H_
#  include <rld.h>
# endif
#endif /* NeXT */
%}

!ObjectFileLoader class methodsFor:'initialization'!

initialize
    "name of object file, where initial symbol table is found"

    mySymbolTable := 'smalltalk'.
    stubNr := 1.
    verbose := false
!

verbose:aBoolean
    "turn on/off debug traces"

    verbose := aBoolean

    "ObjectFileLoader verbose:true"
! !

!ObjectFileLoader class methodsFor:'command defaults'!

needSeparateIDSpaces
    "return true, if we need separate I and D spaces"

    |os cpu|

    os := OperatingSystem getSystemType.
    cpu := OperatingSystem getCPUType.

    (os = 'sunos') ifTrue:[
        (cpu = 'sparc') ifTrue:[ ^ true ]
    ].
    self error:'dont know if we need sepId'
!

absLd:file text:textAddr data:dataAddr
   "this should return a string to link file.o to absolute address"

    |os cpu|

    os := OperatingSystem getSystemType.
    cpu := OperatingSystem getCPUType.
    (os = 'sunos') ifTrue:[
        (cpu = 'sparc') ifTrue:[
"
            ^ ('ld -A ' , mySymbolTable , ' -x -Bstatic -Ttext '
               , (textAddr printStringRadix:16) , ' -Tdata '
               , (dataAddr printStringRadix:16) , ' ' , file)
"
            ^ ('ld -A ' , mySymbolTable , ' -T ',
                          (textAddr printStringRadix:16),
                          ' -N -x ' , file)

        ]
    ].
"
    (os = 'ultrix') ifTrue:[
        (cpu = 'mips') ifTrue:[
            ^ ('ld -A ' , mySymbolTable , ' -x -N -T ' , (textAddr printStringRadix:16) , ' ' , file)
        ]
    ].
"
    self error:'do not know how to link absolute'
! !

!ObjectFileLoader class methodsFor:'dynamic loading'!

loadFile:aFileName library:librariesString withBindings:bindings in:aClass
    "first, load the file itself"

    (self loadFile:aFileName with:librariesString) ifFalse:[^ false].

    "then, create stubs"
    self bindExternalFunctions:bindings in:aClass
!

loadFile:aFileName withBindings:bindings in:aClass
    "load an object file containing external functions, and bind the functions as described 
     in bindings, which is an Array of
        (selector functionName argTypes returnType)
     entries, example:
     #(
        (sel1:and: 'f1' (SmallInteger SmallInteger)    nil)   -> bind 'aClass sel1:and:' to: 'void f1(int, int)'
        (sel2:and: 'f2' (String SmallInteger)       String)   -> bind 'aClass sel2:and:' to: 'char *f2(char *, int)'
      )
    "

    "first, load the file itself"

    (self loadFile:aFileName) ifFalse:[^ false].

    "then, create stubs"
    self bindExternalFunctions:bindings in:aClass
!

bindExternalFunctions:bindings in:aClass
    | selector functionName argTypes returnType allOk |

    allOk := true.
    bindings do:[:aBinding |
        selector := aBinding at:1.
        functionName := aBinding at:2.
        argTypes := aBinding at:3.
        returnType := aBinding at:4.
        (self createStubFor:selector calling:functionName args:argTypes returning:returnType in:aClass)
        isNil ifTrue:[
            Transcript showCr:'binding of ' , functionName , ' failed.'.
            allOk := false
        ]
    ].
    ^ allOk
! !

!ObjectFileLoader class methodsFor:'creating stubs'!

storeGlobalAddressesOn:aStream

    Smalltalk allKeysDo:[:key |
        self storeGlobalAddressOf:key on:aStream
    ]

    "ObjectFileLoader storeGlobalAddressesOn:Transcript"
    "|f|
     f := FileStream newFileNamed:'syms.c'.
     ObjectFileLoader storeGlobalAddressesOn:f.
     f close"
!

storeGlobalAddressOf:aSymbol on:aStream
    |globalName|

    globalName := aSymbol asString.
    (globalName includes:$:) ifTrue:[
        globalName replaceAll:$: by:$_
    ].

    aStream nextPutAll:'#define ',globalName,'_addr '.
    aStream nextPutAll:(Smalltalk cellAt:aSymbol) printString.
    aStream cr.

    aStream nextPutAll:'#define ',globalName,' ( *( (OBJ *) ',globalName,'_addr))'.
    aStream cr

    "ObjectFileLoader storeGlobalAddressOf:#String on:Transcript"
    "ObjectFileLoader storeGlobalAddressOf:#Symbol on:Transcript"
!

createStubFor:aSelector calling:functionName args:argTypes returning:returnType in:aClass
    "create a method calling a stub function"

    |address newMethod|

    address := self createStubCalling:functionName args:argTypes returning:returnType.
    address isNil ifTrue:[^ nil].

    newMethod := Method new.
    newMethod code:address.
    newMethod category:'external functions'.
    newMethod numberOfMethodVars:0.
    newMethod stackSize:0.

    aClass class addSelector:aSelector withMethod:newMethod.

    SilentLoading ifFalse:[
        Transcript showCr:('created stub: ',aClass class name,' ', aSelector)
    ].

    ^ newMethod

    "ObjectFileLoader createStubFor:#printf: calling:'printf' args:#(String) returning:nil in:TestClass"
    "ObjectFileLoader createStubFor:#printf:with: calling:'printf' 
                                            args:#(String SmallInteger) returning:nil in:TestClass"
!

createStubCalling:functionName args:argTypes returning:returnType
    "create a stub function for calling functionName - return the address of the
     function in core or nil on error"

    |baseName p t l handle address stubName|

    stubName := 'stub000' , (stubNr printStringRadix:16).
    stubName := stubName copyFrom:(stubName size - 7) to:(stubName size).

    baseName := self createStubSource:stubName calling:functionName args:argTypes returning:returnType.
    baseName isNil ifTrue:[^ nil].

    "compile it ..."
    verbose ifTrue:[
        Transcript showCr:'compiling stub ...', baseName. Transcript endEntry
    ].

    (OperatingSystem executeCommand:('make /tmp/' , baseName , '.o')) ifFalse:[
        Transcript showCr:'compilation error.'.
        ^ nil
    ].
    OperatingSystem executeCommand:('mv ' , baseName , '.o /tmp/' , baseName , '.o').
    verbose ifFalse:[
        OperatingSystem executeCommand:('rm /tmp/' , baseName , '.c').
    ].

    (OperatingSystem getOSType = 'sys5.4') ifTrue:[
        "make it a sharable object"

        verbose ifTrue:[
            Transcript showCr:'makeing shared object stub ...', baseName. Transcript endEntry.
        ].
        OperatingSystem executeCommand:('ld -G -o /tmp/',baseName,'.so /tmp/',baseName,'.o').

        "attach to it"
        handle := self openDynamicObject:('/tmp/',baseName,'.so').
        handle isNil ifTrue:[
            Transcript showCr:('dlopen error:', '/tmp/',baseName,'.so').
            ^ nil
        ].
        "find the stubs address"
        address := self getSymbol:stubName from:handle.
        address isNil ifTrue:[
            Transcript showCr:'dlsym failed'.
             ^ nil
        ]
    ].

    (OperatingSystem getOSType = 'sunos') ifTrue:[
        "load it"
        (self loadFile:('/tmp/' , baseName , '.o')) ifFalse:[
            Transcript showCr:'load error.'.
            ^ nil
        ].

        "find the stubs address (use nm to get the address)"
        t := Text new.
        p := PipeStream readingFrom:('nm SymbolTable|grep ' , stubName , ' |grep T').
        [p atEnd] whileFalse:[
            l := p nextLine.
            l notNil ifTrue:[
                t add:l
            ]
        ].
        p close.
        (t size == 1) ifFalse:[
            Transcript showCr:('oops, ' , stubName , ' not in name-list.').
            ^ nil
        ].
        address := Integer readFrom:(ReadStream on:(t at:1)) radix:16
    ].

    address isNil ifTrue:[
        Transcript showCr:'no way to dynamically load objects'.
        ^ nil
    ].

    verbose ifTrue:[
        Transcript show:'stub ' , stubName , ' address:'.
        Transcript showCr:(address printStringRadix:16).
    ].

    stubNr := stubNr + 1.
    ^ address

    "ObjectFileLoader createStubCalling:'printf' 
                                   args:#(String)
                              returning:nil"
!

createStubSource:stubName calling:functionName args:argTypes returning:returnType
    "create a temp file with stub-code - return base-filename or nil"

    |pid baseName index aStream argName|

    pid := OperatingSystem getProcessId printString.
    baseName := 'stc' ,  pid.
    aStream := FileStream newFileNamed:('/tmp/' , baseName , '.c').
    aStream nextPutAll:'
#include <stc.h>
'.

    OperatingSystem getOSType = 'sys5.4' ifTrue:[
        self storeGlobalAddressesOn:aStream.
    ].

    aStream nextPutAll:'
' , stubName , '(self, __sel,
#ifndef THIS_CONTEXT
    __sender,
#else
# define __sender __thisContext
#endif
    __srch, __pI,
#ifdef PASS_ARG_REF
    __args)
    OBJ __args[];
# define __a1 __args[0]
# define __a2 __args[1]
# define __a3 __args[2]
# define __a4 __args[3]
# define __a5 __args[4]
# define __a6 __args[5]
# define __a7 __args[6]
# define __a8 __args[7]

#else
    __a1, __a2, __a3, __a4, __a5, __a6, __a7, __a8)
    OBJ __a1, __a2, __a3, __a4, __a5, __a6, __a7, __a8;
#endif
    OBJ __sel, __srch;
{
    extern OBJ ByteArray, ExternalStream;
    extern OBJ _ISKINDOF_();
'.

    returnType notNil ifTrue:[
        ((returnType == #SmallInteger) or:[returnType == #Boolean]) ifTrue:[
            aStream nextPutAll:'    int __ret;'
        ] ifFalse:[
            (returnType == #Float) ifTrue:[
                aStream nextPutAll:'    double __ret;'
            ] ifFalse:[
                (returnType == #String) ifTrue:[
                    aStream nextPutAll:'    char *__ret;'
                ] ifFalse:[
                    self error:'returnType ' , returnType, ' not supported'.
                    ^ nil
                ]
            ]
        ].
        aStream cr
    ].

    "gen type checking code"
    argTypes notNil ifTrue:[
        index := 0.
        argTypes do:[:argType |
            (index + 1) timesRepeat:[ aStream nextPutAll:'    '].
            argName := '__a' , (index + 1) printString.

            (argType == #SmallInteger) ifTrue:[
                aStream nextPutAll:'if (_isSmallInteger(' , argName , ')) {'
            ] ifFalse:[
                (argType == #Float) ifTrue:[
                    aStream nextPutAll:'if (_isFloat(' , argName , ')) {'
                ] ifFalse:[
                    (argType == #String) ifTrue:[
                        aStream nextPutAll:'if (_isString(' , argName , ')) {'
                    ] ifFalse:[
                        (argType == #Boolean) ifTrue:[
                            aStream nextPutAll:'if ((' , argName , '==true)'.
                            aStream nextPutAll:'||(' , argName , '==false)) {'
                        ] ifFalse:[
                            (argType == #ByteArray) ifTrue:[
                                aStream nextPutAll:'if (_Class(' , argName , ')==ByteArray) {'
                            ] ifFalse:[
                                (argType == #ExternalStream) ifTrue:[
                                    aStream nextPutAll:'if (_ISKINDOF_(' , argName , ',
#ifndef THIS_CONTEXT
__sender,
#endif
#ifdef PASS_ARG_REF
 &ExternalStream
#else
 ExternalStream
#endif
)==true) {'
                                ] ifFalse:[
                                    self error:'argType ' , argType, ' not supported'.
                                    ^ nil
                                ]
                            ]
                        ]
                    ]
                ]
            ].
            aStream cr.
            index := index + 1
        ]
    ].
    "call the function"

    (index + 1) timesRepeat:[ aStream nextPutAll:'    '].
    returnType notNil ifTrue:[
        aStream nextPutAll:'__ret = '
    ].
    aStream nextPutAll:functionName , '('.
    argTypes notNil ifTrue:[
        index := 0.
        argTypes do:[:argType |
            argName := '__a' , (index + 1) printString.
            (argType == #SmallInteger) ifTrue:[
                aStream nextPutAll:'_intVal(' , argName , ')'
            ] ifFalse:[
                (argType == #Float) ifTrue:[
                    aStream nextPutAll:'_floatVal(' , argName , ')'
                ] ifFalse:[
                    (argType == #String) ifTrue:[
                        aStream nextPutAll:'_stringVal(' , argName , ')'
                    ] ifFalse:[
                        (argType == #Boolean) ifTrue:[
                            aStream nextPutAll:'((' , argName , '==true) ? 1 : 0)'
                        ] ifFalse:[
                            (argType == #ByteArray) ifTrue:[
                                aStream nextPutAll:'(_ByteArrayInstPtr(' , argName , ')->ba_element)'
                            ] ifFalse:[
                                (argType == #ExternalStream) ifTrue:[
                                    aStream nextPutAll:'_intVal(_InstPtr(' , argName , ')->i_instvars[',
                                                       ((ExternalStream allInstVarNames indexOf:'filePointer') - 1) printString, '])'
                                ]
                            ]
                        ]
                    ]
                ]
            ].
            index := index + 1.
            (index == argTypes size) ifFalse:[
                aStream nextPutAll:','
            ]
        ]
    ].
    aStream nextPutAll:');'. aStream cr.

    argTypes notNil ifTrue:[
        argTypes size timesRepeat:[
            index timesRepeat:[ aStream nextPutAll:'    '].
            aStream nextPutAll:'}'. aStream cr.
            index := index - 1
        ]
    ].

    returnType notNil ifTrue:[
        (returnType == #SmallInteger) ifTrue:[
            aStream nextPutAll:'    return _MKSMALLINT(__ret);'
        ] ifFalse:[
            (returnType == #Float) ifTrue:[
                aStream nextPutAll:'    return _MKFLOAT(__ret, __s);'
            ] ifFalse:[
                (returnType == #String) ifTrue:[
                    aStream nextPutAll:'    return (__ret ? _MKSTRING(__ret, __s) : nil);'
                ] ifFalse:[
                    (returnType == #Boolean) ifTrue:[
                        aStream nextPutAll:'    return (__ret ? true : false);'
                    ]
                ]
            ]
        ]
    ] ifFalse:[
        aStream nextPutAll:'    return self;'
    ].
    aStream cr.

    aStream nextPutAll:'}'. aStream cr.
    aStream close.
    ^ baseName

    "ObjectFileLoader createStubSource:'stub1' calling:'printMessage'  args:#(String) returning:nil"
    "ObjectFileLoader createStubSource:'stub2' calling:'printMessage2' args:#(String SmallInteger) returning:#String"
    "ObjectFileLoader createStubSource:'stub3' calling:'sqrt'          args:#(Float) returning:#Float"
    "ObjectFileLoader createStubSource:'stub4' calling:'checking'      args:#(SmallInteger SmallInteger) returning:#Boolean"
    "ObjectFileLoader createStubSource:'stub5' calling:'fprintf'       args:#(ExternalStream  String) returning:#SmallInteger"

! !

!ObjectFileLoader class methodsFor:'loading objects'!

loadFile:oFile with:librariesString
    "load in an object files code, linking in libraries"

    |tmpOfile errStream errors errText ok pid|

    pid := OperatingSystem getProcessId printString.
    tmpOfile := '/tmp/stc_ld' ,  pid.
    verbose ifTrue:[
        Transcript showCr:'executing: ' , ('ld -o ',tmpOfile,' -r ' , oFile , ' ' , librariesString , '>/tmp/out 2>/tmp/err')
    ].
    (OperatingSystem executeCommand:'ld -o ',tmpOfile,' -r ' , oFile , ' ' , librariesString , '>/tmp/out 2>/tmp/err')
    ifFalse:[
        errStream := FileStream oldFileNamed:'/tmp/err'.
        errStream isNil ifTrue:[
            self notify:'errors during link.'
        ] ifFalse:[
            errors := errStream contents.
            errText := errors asText.
            (errText size > 20) ifTrue:[
                errText grow:20.
                errText add:'... '.
                errors := errText
            ].
            OperatingSystem executeCommand:'rm /tmp/err /tmp/out'.
            self notify:('link errors:\\' , errors asString) withCRs
        ].
        ^ false
    ].
    ok := self loadFile:tmpOfile.
    OperatingSystem executeCommand:('rm ' , tmpOfile).
    ^ ok
!

loadFile:oFile
    "load in an object files code"

    | unixCommand errStream errors errText
      text data textSize dataSize |

    "find out, how much memory we need"

    textSize := ObjectFile textSizeOf:oFile.
    textSize isNil ifTrue:[
        Transcript showCr:'bad text-size in object file'.
        ^ false
    ].
    verbose ifTrue:[
        Transcript showCr:'text-size: ' , (textSize printStringRadix:16)
    ].

    dataSize := ObjectFile dataSizeOf:oFile.
    dataSize isNil ifTrue:[
        Transcript showCr:'bad data-size in object file'.
        ^ false
    ].

    verbose ifTrue:[
        Transcript showCr:'data-size: ' , (dataSize printStringRadix:16)
    ].

    "allocate some memory for text and some for data;
     then call linker to link the file to those addresses"

    self needSeparateIDSpaces ifTrue:[
        text := ExternalBytes newForText:textSize.
        (dataSize ~~ 0) ifTrue:[
            data := ExternalBytes newForData:dataSize
        ].

        text isNil ifTrue:[
            Transcript showCr:'cannot allocate memory for text'.
            ^ false
        ].

        verbose ifTrue:[
            Transcript showCr:'text: ' , (text address printStringRadix:16)
        ].

        (dataSize ~~ 0) ifTrue:[
            (data isNil) ifTrue:[
                Transcript showCr:'cannot allocate memory for data'.
                text notNil ifTrue:[text free].
                ^ false
            ].
            verbose ifTrue:[
                Transcript showCr:'data: ' , (data address printStringRadix:16)
            ]
        ]

    ] ifFalse:[
        text := ExternalBytes newForText:(textSize + dataSize).
        text isNil ifTrue:[
            Transcript showCr:'cannot allocate memory for text'.
            ^ false
        ].
        verbose ifTrue:[
            Transcript showCr:'addr: ' , (text address printStringRadix:16)
        ]
    ].

    unixCommand := (self absLd:oFile text:text address data:data address) , ' >/tmp/out 2>/tmp/err'.

    verbose ifTrue:[
        Transcript showCr:'executing: ' , unixCommand
    ].

    Transcript showCr:'linking ...'.
    (OperatingSystem executeCommand:unixCommand) ifFalse: [
        errStream := FileStream oldFileNamed:'/tmp/err'.
        errStream notNil ifTrue:[
            errors := errStream contents.
            errText := errors asText.
            (errText size > 20) ifTrue:[
                errText grow:20.
                errText add:'... '.
                errors := errText
            ].
            OperatingSystem executeCommand:'rm /tmp/err /tmp/out'.
            self notify:('link errors:\\' , errors asString) withCRs
        ].
        Transcript showCr:'link unsuccessful.'.
        text notNil ifTrue:[text free].
        data notNil ifTrue:[data free].
        ^ false
    ].

    Transcript showCr:'link successful'.

    OperatingSystem executeCommand:'rm /tmp/err /tmp/out'.

    "only thing left to do is to load in text at textAddr and
     data at dataAddr ... "

    (ObjectFile loadObjectFile:'a.out'
                textAddr:text address textSize:textSize
                dataAddr:data address dataSize:dataSize) isNil ifTrue: [
        Transcript showCr:'load in error'.
        text notNil ifTrue:[text free].
        data notNil ifTrue:[data free].
        ^ false
    ].

    Transcript showCr:'load in successful'.

    OperatingSystem executeCommand:'mv a.out SymbolTable'.
    mySymbolTable := 'SymbolTable'.
    ^ true
! !

!ObjectFileLoader class methodsFor:'dynamic class loading'!

loadClass:aClassName fromObjectFile:aFileName
    "load a compiled class (.o-file) into the image"

    |handle initAddr symName|

    handle := self openDynamicObject:aFileName.
    handle isNil ifTrue:[
        Transcript showCr:('openDynamic: ',aFileName,' failed.').
        ^ nil
    ].
    OperatingSystem getOSType = 'sys5.4' ifTrue:[
        symName := '_' , aClassName , '_Init'
    ] ifFalse:[
        symName := '__' , aClassName , '_Init'
    ].
    initAddr := self getSymbol:symName from:handle.
    initAddr isNil ifTrue:[
        Transcript showCr:('no symbol: ',symName,' in ',aFileName).
        ^ nil
    ].
    self callFunctionAt:initAddr.
    ^ Smalltalk at:aClassName asSymbol

    "ObjectFileLoader loadClass:'Tetris'      fromObjectFile:'../clients/Tetris/Tetris.o'"
    "ObjectFileLoader loadClass:'TetrisBlock' fromObjectFile:'../clients/Tetris/TBlock.o'"
!

loadObjectFile:aFileName
    "load a compiled class (.o-file) into the image; the class name
     is not needed (multiple definitions may be in the file)"

    |handle initAddr symName className|

    handle := self openDynamicObject:aFileName.
    handle isNil ifTrue:[
        Transcript showCr:('openDynamic: ',aFileName,' failed.').
        ^ nil
    ].
    className := OperatingSystem baseNameOf:aFileName.
    (className endsWith:'.o') ifTrue:[
        className := className copyFrom:1 to:(className size - 2)
    ].
    OperatingSystem getOSType = 'sys5.4' ifTrue:[
        symName := '_' , className , '_Init'
    ] ifFalse:[
        symName := '__' , className , '_Init'
    ].
    initAddr := self getSymbol:symName from:handle.
    initAddr isNil ifTrue:[
        Transcript showCr:('no symbol: ',symName,' in ',aFileName).
        ^ nil
    ].
    self callFunctionAt:initAddr.
    ^ self
! !

!ObjectFileLoader class methodsFor:'dynamic object access'!

openDynamicObject:pathName
    "open an object-file (map into my address space).
     Return a non-nil handle if ok, nil otherwise.
     This function is not supported on all architectures."

    |low hi|
%{
#ifdef SYSV4
#   include <dlfcn.h>
    void *handle;

    if ((pathName == nil) || _isString(pathName)) {
        if (pathName == nil)
            handle = dlopen((char *)0, RTLD_NOW);
        else
            handle = dlopen(_stringVal(pathName), RTLD_NOW);
        if (handle) {
            printf("open %s handle = %x\n", _stringVal(pathName), handle);
            low = _MKSMALLINT( (int)handle & 0xFFFF );
            hi = _MKSMALLINT( ((int)handle >> 16) & 0xFFFF );
        } else {
            printf("dlopen %s error: <%s>\n", _stringVal(pathName), dlerror());
        }
    }
#endif
#ifdef NeXT
    long result;
    char *files[2];
    NXStream *errOut;

    if (_isString(pathName)) {
        files[0] = _stringVal(pathName);
        files[1] = (char *)0;
        errOut = NXOpenFile(2, 2);
        result = rld_load(errOut,
                          (struct mach_header **)0,
                          files,
                          (char *)0);
        NXClose(errOut);
        if (result) {
            printf("rld_load %s ok\n", _stringVal(pathName));
            RETURN ( _MKSMALLINT(1) ); /* a dummy handle */
        }
    }
    RETURN ( nil );
#endif
%}
.
    low notNil ifTrue:[
        ^ (hi * 16r10000) + low
    ].

    ^ nil

    "sys5.4:
     |handle|
     handle := ObjectFileLoader openDynamicObject:'../stc/mod1.so'.
     ObjectFileLoader getSymbol:'module1' from:handle"
    "next:
     |handle|
     handle := ObjectFileLoader openDynamicObject:'../goodies/Path/AbstrPath.o'.
     ObjectFileLoader getSymbol:'__AbstractPath_Init' from:handle"
!

closeDynamicObject:handle
    "close an object-file (unmap from my address space)."

    |low hi|

    hi := handle // 16r10000.
    low := handle \\ 16r10000.
%{
#ifdef SYSV4
#   include <dlfcn.h>
    void *h;
    int val;

    if (_isSmallInteger(low) && _isSmallInteger(hi)) {
        val = (_intVal(hi) << 16) + _intVal(low);
        h = (void *)(val);
        printf("close handle = %x\n", h);
        dlclose(h);
    }
#endif
%}
!

getSymbol:aString from:handle
    "return the address of a symbol from a dynamically loaded object file.
     Handle must be the one returned previously from openDynamicObject.
     Return the address of the symbol, or nil on any error."

    |low hi lowAddr hiAddr|

    hi := handle // 16r10000.
    low := handle \\ 16r10000.
%{
#ifdef SYSV4
#   include <dlfcn.h>
    void *h;
    void *addr;
    int val;

    if (_isSmallInteger(low) && _isSmallInteger(hi)) {
        val = (_intVal(hi) << 16) + _intVal(low);
        h = (void *)(val);
        if (_isString(aString)) {
            printf("get sym <%s> handle = %x\n", _stringVal(aString), h);
            addr = dlsym(h, _stringVal(aString));
            if (addr) {
                printf("addr = %x\n", addr);
                lowAddr = _MKSMALLINT( (int)addr & 0xFFFF );
                hiAddr = _MKSMALLINT( ((int)addr >> 16) & 0xFFFF );
            } else {
                printf("dlsym %s error: <%s>\n", _stringVal(aString), dlerror());
            }
        }
    }
#endif
#ifdef NeXT
    unsigned long addr;
    long result;
    NXStream *errOut;

    if (_isString(aString)) {
        printf("get sym <%s>\n", _stringVal(aString));
        errOut = NXOpenFile(2, 2);
        result = rld_lookup(errOut,
                            _stringVal(aString),
                            &addr);
        NXClose(errOut);
        if (result) {
            printf("addr = %x\n", addr);
            lowAddr = _MKSMALLINT( (int)addr & 0xFFFF );
            hiAddr = _MKSMALLINT( ((int)addr >> 16) & 0xFFFF );
        }
    }
#endif
%}
.
    lowAddr notNil ifTrue:[
        ^ (hiAddr * 16r10000) + lowAddr
    ].
    ^ nil
!

releaseSymbolTable
    "this is needed on NeXT to forget loaded names. If this wasnt done,
     the same class could nat be loaded in again due to multiple defines.
     On other architectures, this is not needed and therefore a noop."

%{
#ifdef NeXT
    NXStream *errOut;

    errOut = NXOpenFile(2, 2);
    rld_unload_all(errOut, (long)0);
    rld_load_basefile(errOut, "smalltalk");
    NXClose(errOut);
#endif
%}
!

callFunctionAt:address
    "call a function at address - this is very dangerous.
     This is needed to call the classes init-function after loading in a
     class-object file. Dont use in your programs."

    |low hi lowAddr hiAddr|

    hi := address // 16r10000.
    low := address \\ 16r10000.
%{
    void (*addr)();
    unsigned val;
    typedef void (*VOIDFUNC)();

    if (_isSmallInteger(low) && _isSmallInteger(hi)) {
        val = (_intVal(hi) << 16) + _intVal(low);
        addr = (VOIDFUNC) val;
        (*addr)();
    }
%}
! !

ObjectFileLoader initialize!
